import os
import subprocess


def docker_lib_path():
    """
    Get the path to the docker lib folder
    :return:
    """
    output = subprocess.check_output(["docker", "info", "--format={{.DockerRootDir}}"]).decode()
    return output


def docker_disk_free():
    """
    Return the amount of space free on docker's main fs in MB
    :return:
    """
    folder = docker_lib_path().strip()
    if os.path.exists(folder):
        if os.path.islink(folder):
            folder = os.readlink(folder)

        statvfs = os.statvfs(folder)
        return int((statvfs.f_frsize * statvfs.f_bavail) / 1024 / 1024)
    return -1




"""
Windows Service Integration
"""
import sys
import win32serviceutil, servicemanager

from DockerScout.win32service.service import DockerScoutService, ensure_wintypes_path


if __name__ == '__main__':
    ensure_wintypes_path()
    #if len(sys.argv) == 1:
    #    servicemanager.Initialize()
    #    servicemanager.PrepareToHostSingle(DockerScoutService)
    #    servicemanager.StartServiceCtrlDispatcher()
    #    return

    win32serviceutil.HandleCommandLine(DockerScoutService, argv=list(sys.argv))



import socket
import os
import sys
import servicemanager
import win32service
import win32serviceutil

from DockerScout.monitor import ImageScanner, run


def ensure_wintypes_path():
    import site
    sites = site.getsitepackages()

    pywin32_system32 = "pywin32_system32"
    for item in sites:
        if item.lower().endswith("site-packages"):
            addpath = os.path.join(item, pywin32_system32)
            if os.path.isdir(addpath):
                append_system_path(addpath)


def append_system_path(addpath):
    # borrowed from https://stackoverflow.com/a/21452408/148415
    import winreg as reg

    # read the value
    # use this if you need to modify the system variable and if you have admin privileges
    key = reg.OpenKey(reg.HKEY_LOCAL_MACHINE, r'SYSTEM\CurrentControlSet\Control\Session Manager\Environment', 0, reg.KEY_ALL_ACCESS)
    value, _ = reg.QueryValueEx(key, 'PATH')

    if addpath in value:
        return

    # modify it
    value = value + os.pathsep + addpath

    # write it back
    reg.SetValueEx(key, 'PATH', 0, reg.REG_EXPAND_SZ, value)
    reg.CloseKey(key)


class DockerScoutService(win32serviceutil.ServiceFramework):
    _svc_name_ = "DockerScout"
    _svc_display_name_ = "Docker Scout"
    _svc_description_ = "Clean up unused docker images"

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        servicemanager.LogInfoMsg("New Instance {}".format(self._svc_display_name_))

        socket.setdefaulttimeout(60)

    def info(self, msg):
        servicemanager.LogInfoMsg(msg)

    def SvcStop(self):
        servicemanager.LogInfoMsg("Stopping {}".format(self._svc_display_name_))
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        self.ReportServiceStatus(win32service.SERVICE_STOPPED)
        sys.exit(0)
        ImageScanner.getinstance().end = True

    def SvcDoRun(self):
        self.info("Starting {}".format(self._svc_display_name_))
        self.ReportServiceStatus(win32service.SERVICE_START_PENDING)
        self.ReportServiceStatus(win32service.SERVICE_RUNNING)
        run([])



import re

time_re = re.compile(r"(\d+)([hmsdw]*)")
space_re = re.compile(r"(\d+)([mg]*)")


def time_suffix(text):
    """
    Return the time string "10s" "3h" as seconds
    :param text:
    :return: number of seconds
    """
    m = re.match(time_re, text)
    multiply = 1
    if len(m.groups()) == 2:
        suffix = m.group(2)

        if suffix == "h":
            multiply = 60 * 60
        if suffix == "m":
            multiply = 60
        if suffix == "d":
            multiply = 60 * 60 * 24
        if suffix == "w":
            multiply = 60 * 60 * 24 * 7
        number = int(m.group(1))
    else:
        raise ValueError("Expected a string with a suffix eg (10s, 3h)")

    return multiply * number


def space_suffix(text):
    """
    Return the disk space string "10g" "100m" in megabytes
    :param text:
    :return:
    """
    m = re.match(space_re, text)
    multiply = 1
    if len(m.groups()) == 2:
        suffix = m.group(2)

        if suffix == "g":
            multiply = 1024

        number = int(m.group(1))
    else:
        raise ValueError("Expected a string with a suffix eg (10g, 300m)")

    return multiply * number



import os
import yaml
import platform

from .utils import space_suffix


CONFIG_DIR = ".docker-scout"
HISTORY_FILE = "images.json"
CONFIG_FILE = "docker-scout.yml"


def config_dir():
    """
    Get the config folder
    :return:
    """
    override = os.getenv("DOCKERSCOUT_DIR", None)
    if override is not None:
        if not os.path.exists(override):
            os.makedirs(override)
        return override

    if platform.system() == "Windows":
        progdata = os.getenv("ProgramData")
        if progdata:
            cfg = os.path.join(progdata, CONFIG_DIR)
    else:
        home = os.getenv("HOME", "/tmp")
        cfg = os.path.join(home, CONFIG_DIR)

    if not os.path.exists(cfg):
        os.makedirs(cfg)
    return cfg


def config_path():
    """
    Get the config file path
    :return:
    """
    # try some system-wide places first
    if platform.system() == "Windows":
        cfg = os.getenv("windir", None)
    if platform.system() == "Linux":
        cfg = "/etc/"
    local_cfg = os.path.join(config_dir(), CONFIG_FILE)
    if not os.path.exists(local_cfg):
        global_cfg = os.path.join(cfg, CONFIG_FILE)
        if os.path.exists(global_cfg):
            return global_cfg

    return os.path.join(cfg, CONFIG_FILE)


class ScoutConfigV1(object):
    DEFAULT_MIN_DISK_FREE = "10g"
    DEFAULT_MIN_DISK_FREE_WINDOWS = "30g"

    def __init__(self, data):
        if data is not None:
            self.data = data["scout"]["v1"]
        else:
            self.data = {"scout": {"v1": {}}}

    def get_defailt_disk_free(self):
        if platform.system() == "Windows":
            return self.DEFAULT_MIN_DISK_FREE_WINDOWS
        return self.DEFAULT_MIN_DISK_FREE

    def min_disk(self, change=None):
        """
        Return the point at which we start cleaning old images
        :return:
        """
        if change is not None:
            self.data["min-disk-free"] = change
        return space_suffix(self.data.get("min-disk-free", self.get_defailt_disk_free()))


def read_config(filename):
    """
    Read the docker scout configuration. If no config exist, use defaults
    :param filename:
    :return:
    """
    if os.path.exists(filename):
        with open(filename, "r") as configfile:
            data = yaml.load(configfile, Loader=yaml.SafeLoader)
    else:
        data = None

    return ScoutConfigV1(data)


def save_config(cfg, filename):
    """
    Save the configuration
    """
    with open(filename, "wb") as configfile:
        configfile.write(yaml.dump(cfg.data, indent=4, encoding="utf-8"))
        


"""
Monitor currently running docker containers so we can keep track of images to keep
"""
import os
import sys
import argparse
import logging
import platform
import subprocess
import time
from threading import Thread
from . import memory, config, utils
from .disk import docker_disk_free


logging.root.setLevel(logging.INFO)
logging.root.name = "docker.scout.monitor"


def get_all_images():
    """
    Get a all images
    :return:
    """
    output = subprocess.check_output(["docker", "image", "ls", "-q", "-a", "--no-trunc", "--format={{.ID}}"]).decode()
    all_images = []
    for line in output.split("\n"):
        all_images.append(line.strip())

    return all_images


def clean(image):
    """
    Remove an image
    :param image:
    :return:
    """
    if image:
        logging.info("removing image {}".format(image))
        try:
            subprocess.check_call(["docker", "image", "rm", image])
            memory.forget(image)
            return True
        except Exception as err:
            logging.error("error in clean {}".format(err))
    return False


def resolve_image(image):
    """
    Get the local ID of an image
    :param image:
    :return:
    """
    try:
        output = subprocess.check_output(["docker", "image", "inspect", "--format", '{{.ID}}', image]).decode()
        return output.strip()
    except Exception as err:
        logging.error("error in resolve_image {}".format(err))
        return None


class ImageScanner(Thread):
    """
    Thread to scan for newly started images and clean up when the disk gets low on space
    """

    instance = None

    @classmethod
    def getinstance(cls):
        return cls.instance

    def __init__(self, touch, minfree):
        super(ImageScanner, self).__init__()
        self.touch = touch
        self.minfree = minfree
        self.end = False
        self.instance = self

    def clean(self):
        """
        If disk space is below the limit, clean the oldest image
        :return:
        """
        space = docker_disk_free()
        if space < self.minfree:
            # only try to delete images not seen running in the last 2 mins
            only_delete_before = time.time() - 120

            logging.info("We are low on space ({}mb < {}mb)".format(space, self.minfree))
            # prune all stopped containers
            logging.info("Pruning stopped containers..")
            subprocess.call(["docker", "container", "prune", "-f"])

            # now clean images until we are above the limit or there are no more images
            current_images = get_all_images()
            history = memory.read_images()
            for image in current_images:
                if image not in history:
                    # we've not seen it run, destroy it
                    logging.info("Cleaning image {} (unseen)".format(image))
                    clean(image)
                if docker_disk_free() > self.minfree:
                    break

            # sort the images by last seen, and delete the oldest first and stop
            # when we delete at least one image
            least_used_first = sorted(history.items(), key=lambda x: x[1])
            for image, lastseen in least_used_first:
                if docker_disk_free() > self.minfree:

                    break
                if lastseen > only_delete_before:
                    break
                logging.info("Cleaning image {} (least used)".format(image))
                clean(image)

            logging.info("Cleaning done. {}Mb free".format(docker_disk_free()))

    def run(self):
        """
        Use 'docker monitor' to watch for containers being started and report the image used
        :return:
        """
        while not self.end:
            logging.info("Starting docker events monitor")
            proc = subprocess.Popen(
                ["docker", "events",
                 "--filter", "Type=container",
                 "--format", "{{.Action}} {{.Actor.Attributes.image}}"],
                stderr=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stdin=subprocess.DEVNULL)

            while proc.poll() is None:
                line = proc.stdout.readline()
                if not line:
                    break
                try:
                    image = line.decode().strip()
                    action, image = image.split(" ", 1)
                    if action != "create":
                        continue
                    logging.debug("{} {}".format(action, image))

                    imageid = resolve_image(image)
                    logging.info("noticed image start {}".format(imageid))
                    if self.touch:
                        self.touch(imageid)

                    self.clean()
                except Exception as err:
                    logging.error("error touching image {}".format(err))

            proc.wait()


SYSTEMD_SERVICE = """
[Unit]
Description=Docker Scout Image Cleaner
After=multi-user.target
After=docker.service

[Service]
Type=simple
Environment=DOCKERSCOUT_DIR=/etc/docker-scout
ExecStart={cmdline}
StandardInput=tty-force
StandardOutput=syslog
StandardError=syslog
SyslogIdentifier=DockerScout

[Install]
WantedBy=multi-user.target

"""


def do_install(configfile):
    """
    Install as a system service
    :param configfile:
    :return:
    """

    cmdline = "{} -m DockerScout --config /etc/docker-scout.yml".format(sys.executable)

    if platform.system() == "Linux":
        with open("/etc/systemd/system/dockerscout.service", "w") as service:
            service.write(SYSTEMD_SERVICE.format(cmdline=cmdline))
        subprocess.check_call(["systemctl", "daemon-reload"])
        subprocess.check_call(["systemctl", "restart", "dockerscout"])

    elif platform.system() == "Windows":
        import site
        # we have to put all the options to "install" before the "install" because it uses getopt()
        cmdline = [sys.executable, "-m", "DockerScout.win32service", "--startup", "auto", "install"]
        subprocess.check_call(cmdline)
    else:
        raise NotImplemented("Service install not supported on {}".format(platform.system()))


def run(args):
    """
    Monitor running images
    :return:
    """
    parser = argparse.ArgumentParser(args)
    parser.add_argument("--config", type=str,
                        metavar="CONFIG",
                        default=config.config_path(),
                        help="Specify a config file")
    parser.add_argument("--install", default=False,
                        action="store_true",
                        help="Install as a system service")

    parser.add_argument("--set-minspace", default=None, type=str,
                        dest="minspace",
                        help="Set a new min space value")                        

    opts = parser.parse_args(args)

    print("config file is {}".format(opts.config))
    print("history file is {}".format(os.path.join(config.config_dir(), config.HISTORY_FILE)))

    settings = config.read_config(opts.config)
    if opts.minspace:
        settings.min_disk()
        config.save_config(settings, opts.config)
        return

    if opts.install:
        do_install(opts.config)
    else:
        logging.info("aiming for {} mb disk free".format(settings.min_disk()))
        mindisk = utils.space_suffix(str(settings.min_disk()))

        scanner = ImageScanner(memory.notice, mindisk)
        scanner.setDaemon(True)
        scanner.start()
        while scanner.is_alive() and not scanner.end:
            scanner.join(10.0)




"""
Remember when we last saw each docker image
"""
import os
import json
import time
from .config import config_dir, HISTORY_FILE


def read_images():
    """
    Read the image history
    :return:
    """
    histfile = os.path.join(config_dir(), HISTORY_FILE)
    if os.path.exists(histfile):
        with open(histfile, "r") as cfg:
            return json.load(cfg)
    return {}


def save_images(images):
    """
    Save the image history
    :return:
    """
    if not os.path.exists(config_dir()):
        os.makedirs(config_dir())

    histfile = os.path.join(config_dir(), HISTORY_FILE)

    with open(histfile, "w") as cfg:
        json.dump(images, cfg, indent=2)


def notice(image):
    """
    Remember that we just saw an image
    :param image:
    :return:
    """
    now = time.time()

    images = read_images()
    images[image] = now
    save_images(images)


def forget(image):
    """
    Forget history of an image
    :param image:
    :return:
    """
    images = read_images()
    if image in images:
        del images[image]
    save_images(images)



import platform
from distutils.core import setup

deps = []
pkgs = ["DockerScout"]
if platform.system() == "Windows":
    deps.append("pywin32")
pkgs.append("DockerScout.win32service")

setup(
    name="dockerscout",
    version="0.0.4",
    description="Service to prevent unused docker images filling your disk",
    author="Ian Norton",
    author_email="inorton@gmail.com",
    url="https://gitlab.com/inorton/docker-scout",
    packages=pkgs,
    install_requires=deps,
    platforms=["any"],
    license="License :: OSI Approved :: MIT License",
    long_description="Service to prevent unused docker images filling your disk by monitoring docker container startup events"
)
